

/* defs.h -- globally used defines for the Crystal Image application */


#ifndef __IMAGER_DEFS_H
#define __IMAGER_DEFS_H


#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>
#include <gdk-pixbuf/gdk-pixbuf.h>

#include <sys/types.h>



//#define PACKAGE "cimage"
//#define VERSION "0.9 Beta"  Note - these are defined in ../config.h


/* If we want to use the DICOM code, define the following symbol: */

#define _DICOM_



/* The only non-standard element is the RGB triplet */

typedef struct
{
        unsigned char red;
        unsigned char green;
        unsigned char blue;
}
rgbtriplet;

/* For Imlib2, we need a 32-bit RGBA type. Now this is not really a triplet, but... */

typedef struct
{
        unsigned char red;
        unsigned char green;
        unsigned char blue;
		unsigned char alpha;
}
rgbatriplet;



/* HSV gets a nice struct as well */

typedef struct
{
        short hue;				/* Range may be -359 to +359 */
        unsigned char sat;
        unsigned char val;
}
hsvtriplet;


/* The result of a lsq fit */


typedef struct
{
	double H;
	double intcpt;
	double r;
}
fitresult;



/*
	The buffer may be any of the following types: char (8-bit unsigned), short
	(16-bit signed), RGB (24-bit coded as 3 chars for each R,G,B), float (23
	bit)
*/

typedef enum { B_CHAR, B_SHORT, B_RGB, B_FLOAT } TYPE_SELECTOR;


/* Interpolation may be performed using the following methods: */

typedef enum { INTP_NN, INTP_BILIN, INTP_BICUB, 
				INTP_SPLINE2, INTP_SPLINE3, INTP_SPLINE4, INTP_SPLINE5 } interpol_method;

/*
	For morphological operations, we define the operations here
*/

typedef enum { M_BYPASS, M_PATCH, M_ERODE, M_SOFTERODE, M_RANK, M_SOFTDILATE, M_DILATE, M_BOUNDERODE, M_CWMF } MORPHOP;


/*
	We have two undo options - stack in memory or on disk 
*/

typedef enum { U_NONE, U_MEMORY, U_DISK } UNDO_TYPE;

/*
	Enumerate file types that can be detected through the header
*/

typedef enum { FTYPE_NONE, FTYPE_TIFFI, FTYPE_TIFFM, FTYPE_VSRT, FTYPE_VFLT, FTYPE_DCM, FTYPE_PGM, FTYPE_PPM } FTYPE;


/* 
	A calibration wedge has the following parameters :
*/

typedef struct
{
	double Dmin;	/* Background intensity */
	double Dmax;	/* Maximum intensity */
	double Mu;		/* Attenuation factor */
	char calibrated; /* Set to 1 if valid calibration available */
}
wedge_data;



/* This is our actual image structure. The structure contains the image
   dimensions in x,y,z; the image type (see enum above); the min and max values
   (float: use maxval.fmax; short and char: use maxval.imax; rgb: also use
   maxval.imax, but contains the max of the triplet); and finally a pointer to
   the data array.
	NOTE that as of version 0.9.0, this structure is self-contained with its
	GUI, and would in principle allow the opening of multiple windows.
*/

typedef struct 
{
		long xmax;
		long ymax; 
		long zmax;
		TYPE_SELECTOR imgtype;
		int uid;
		union { int imax; float fmax; } maxval;	/* for scaling purposes */
		union { int imin; float fmin; } minval; /* handle char and int identical */
		void* data;
		GtkWidget *parent;						/* Pointer to parent window */
		GtkWidget *drawable;					/* Pointer to drawing area */
		GdkPixbuf *pixbuf;						/* gdk_pixbuf link to the drawable */
		int width, height;						/* gdk pixbuf allocation size */
}
image_type;



/*
	Different thresholding functions
*/

typedef enum { TH_BINARY, TH_SOFT, TH_HARD, TH_SOFTSYM, TH_HARDSYM, TH_SATURATE, TH_SATSYM } THRESHOLD_TYPE;


typedef struct
{
	int r;
	int g;
	int b;
	float t;
	float u;
	THRESHOLD_TYPE thresh_type;
	char success;
	image_type* img;		/* Image to operate the threshold on */
}
thresh_struct;

/* The following is used in the findcluster algorithms */

typedef struct
{
	int size;
	long cog_x;
	long cog_y;
	long cog_z;
	int bbox_x1;
	int bbox_y1;
	int bbox_z1;
	int bbox_x2;
	int bbox_y2;
	int bbox_z2;
	int size1;
	int size2;
	int flags;
	float compactness;	/* Compactness of cluster */
	float aspect;		/* Aspect ratio of cluster */
	float irreg;		/* Irregularity (cv) of boundary */
} 
cluster_type;


typedef struct
{
	int x;
	int y;
}
vertex;


typedef struct
{
	float x;
	float y;
}
vertexf;


typedef struct
{
	int x;
	int y;
	int z;
}
vertex3d;

/* This type is primarily used in statistics/measure functions */

typedef struct
{
	long n;
	double Sx;
	double Sxx;
	double SxR, SxG, SxB, SxxR, SxxG, SxxB;
}
stat_type;



#define ROI_MAXVERTICES	1000
#define MAXPOINTSONARAY 600



typedef enum { ROI_NONE, ROI_RECT, ROI_POLY, ROI_CIRCLE } roi_enum;


typedef struct
{
	roi_enum roi_mode; /* 0: no ROI; 1=rect; 2=polygon; 3=circular polygon */
	int bbox_x1;
	int bbox_x2;
	int bbox_y1;
	int bbox_y2;
	int numvertices;
	int radius,xc,yc;
	int xi[ROI_MAXVERTICES];	/* Wastes some memory and limits polygons to 1000 vertices */
	int yi[ROI_MAXVERTICES];	/* but makes life dramatically easier */
	int centfromspider_x;
	int centfromspider_y;
}
roi_type;


/* This structure is used for the spider op segmentation and related algorithms */

typedef struct  
{
	float endpoints[2][2];
	int startx;
	int starty;
	int endx;
	int endy;
	int numPixels;		/* length of ray */
	int XPixelCoords[MAXPOINTSONARAY];
	int YPixelCoords[MAXPOINTSONARAY];
	float pixelValues[MAXPOINTSONARAY]; /* sampled gray values along ray trace */
	float vtxRadialPosition;
	int vtxXPixelCoord;	  /* So these will change based on the edge detection mechanism */
	int vtxYPixelCoord;
	int maxXPixelCoord;	  /* these hold the closest local maximum */
	int maxYPixelCoord;
	int vtxPixelIndex;
	int maxPixelIndex;
	float vtxPixelValue;	/* Pixel value at the intersection */
	float maxPixelValue;	/* Pixel value at local max */
	float vtxOutsideValue;	/* avg value from intersection to end of ray */
	int validVertex ;	  /*  IF zero, include this vertex.  If 1 do not */
	float angleinrads;	  /* angle of a specific ray */
	float rotmatrix[2][2]; 
	float slope;
	float normalslope;
} ray;



/* OBSOLETE -- To complete the information, here is the structure of a 
	GdkImlibImage, copied from gdk_imlib_types.h, just for info purposes,
	thus ifdef'd away. DON'T ENABLE IT	*/
	
#ifndef TRUE
struct _GdkImlibImage
{
    gint             rgb_width, rgb_height;
    unsigned char      *rgb_data;
    unsigned char      *alpha_data;
    gchar              *filename;
	/* the below information is private */
    gint                width, height;
    GdkImlibColor       shape_color;
    GdkImlibBorder      border;
    GdkPixmap          *pixmap;
    GdkBitmap          *shape_mask;
    gchar               cache;
    GdkImlibColorModifier mod, rmod, gmod, bmod;
    GdkImlibModifierMap *map;
    int                 ref_count;
    GHashTable          *pixmaps;
};
#endif



/*====================================================

	G L O B A L   V A R I A B L E S 

======================================================*/


unsigned char version_string[256];		/* Version displayed in the control window */

/* Let's have one image that can be globally accessed */

image_type mainimg;
image_type backupimg;
char mainimg_name[255];

/* ... one image buffer for the secondary display ... */

image_type s_img;

/* ... image(s) and other variables for the undo functions ... */

#define UNDO_MAX	10		/* Maximum of 10 undo levels on disk */
image_type undo_mem;		/* One backup image in memory */
char undo_path[1024];		/* Where to store disk backup images */
UNDO_TYPE undo_function;	/* Undo on disk or in memory */
int undo_level;				/* How many images we have on disk */
int undo_cnt;				/* Number of undo image */
char **undo_stack;			/* file names of undo images */
pid_t cimage_pid;			/* Process ID of this Crystal Image instance */


/* ... and one imlib image and some associated variables */

int autosize;		/* DEPRECATED: display fixed size or scale with display size */
int zoomlevel_x;	/* Now replaced by zoom level. 0=1:1, <0 means shrink, >0 enlarge */
int zoomlevel_y;		/* Allows anisotropic scaling */
int autoscale;		/* Never used: expand contrast or display 1:1 gray values */
int master_abort;	/* Hooked up to the "stop" button to stop lengthy processes - not yet fully implemented */
int master_enable;	/* To disable the menu while a task is running */
int file_ok;		/* Set to 1 if a file is in memory */
int cur_slice;		/* Slice that is currently visible */
int next_uid;		/* Unique ID for each allocated image */
int debuglevel;		/* Tells how much debug junk we want to see in the console */
int skip_license;	/* Flag that suppresses the license window at start */
char tiffreport;	/* Use to disable the import report */
float tiffsave_dpi;	/* value to set the dpi tag to when saving tiffs */
int stat_window_exists;
interpol_method intp_pref;		/* Preferred interpolation method */
int spl_cache_invalidate;		/* Sline interpolation: Cache validation flag */
GdkCursor* crosshair_cursor;	/* over display windows */
GdkCursor* grab_cursor;			/* When grabbing a vertex */
char browspath[1024];			/* Path to browser, for help files */
char helppath[1024];			/* Path to help files */
char dsopath[1024];				/* Path to DSO files */
char imager_sv_path[1024];		/* Path to external standalone 3D viewer */

void (*zslice_callback) (void); /* A function that gets called whenever
									the z-slice button has been changed */

/* LUTs for red, green, blue (display only!) */

unsigned char redlut[256];
unsigned char greenlut[256];
unsigned char bluelut[256];
unsigned char alphalut[256];
unsigned char use_lut;
unsigned char active_cmap;		/* Colormap currently in use (if any) */
unsigned char cmap_window_open;
unsigned char cmap_use_contours;	/* Whether to display contour lines */
float viz_bright, viz_contrast, viz_gamma;


#define CLICKSTACK_MAX	5
#define MAXVERTICES 	360

vertex3d clickstack[CLICKSTACK_MAX];	/* Recent click positions */
int clickstack_last;	/* Pointer to the last entry of the click stack */
int display_clickstack;	/* Display segments of the click stack marching-ants style */

//image_type *s_im;
GtkWidget *secondaryimg;	/* secondary image drawing area */
GtkWidget *secondwnd;		/* secondary image toplevel window */
int secw, sech;				/* Secondary image size */
int	secondaryimage_exists;	/* flag for the "close" operation */
int graphwindow_exists;		/* as above, but for graph display */
image_type tn_img;
int have_thumbnail;

/* Rays from SpiderOp */

ray rayforspider[MAXVERTICES];
ray raystoconnectvtx[MAXVERTICES];
ray rayforvertices[MAXVERTICES];
int numspiderlegs;
int numvertices;
int spider_active;			/* Set to TRUE when the spider window is open */
int snake_active;			/* Set to TRUE when the snake window is open */

/* Leftover from bone densitometry */

wedge_data wedgecalib;


/* Macro interpreter */

int macrorun;	/* Flag indicates that a macro is running */
float mac_user[100];	/* Macro "user" data array */



/* Some macro definitions and mini-functions that I commonly use */

/* ehmm... seems that glib defines these, so double-check that
   the definitions are what we expect */

#ifndef MIN
#define MIN(X,Y)	((X)>(Y)? (Y): (X))
#endif

#ifndef MAX
#define MAX(X,Y)	((X)<(Y)? (Y): (X))
#endif

#ifndef CLAMP
#define CLAMP(X,LOW, HIGH)			( (X)<(LOW)? (LOW) : ( (X)>(HIGH)? (HIGH) : (X) ))
#endif

#define SIN(X)		(sin((X)*1.7453295199e-2))
#define COS(X)		(cos((X)*1.7453295199e-2))

#define SIGN(X)		((X)<0? -1 : 1)		/* simplified sign, zero counts as positive */
#define IROUND(X)	(int)(0.4999999+(X))

#define SQR(X)		((X)*(X))
#define CUB(X)		((X)*(X)*(X))
#define SAFELOG10(X)	((X)<1e-50? -50: log10(X))
#define SAFELOG(X)	((X)<1e-50? -115.13: log(X))

#define ARITHM(X,Y)		((X)*(X)+(Y)*(Y))

#define SDEV(SX,SXX,N)	((N)>1? sqrt(((SXX)-((SX)*(SX)/((N))))/((N)-1)): 0)

/* We count any floating-point number below 1e-15 as zero. This macro can be used to query */
/* TODO: There is a better way now, in imgbuf.c the function nonzero (img,x,y,z) */

#define F_ISZERO(F)		( fabs(F) < 1e-15? 1 : 0 )

#define DATALEN(T)		( (T)==B_SHORT? 2: (T)==B_FLOAT? 4: (T)==B_RGB? 3: 1 )

/* Next we define the weight factors for conversion of RGB to gray.
   We use the standard YUV system factors here. In fact, a simple
   extension to a 3x3 matrix would give the complete YUV conversion.
   Note the sum of all three weights MUST be unity */

#define REDWEIGHT	0.299
#define GREENWEIGHT	0.587
#define BLUEWEIGHT	0.114

#define PI			3.141592654
#define TWOPI		6.283185307
#define SQ2			1.414213562

/* Use float or double for matrix computations... */

#define MFLOAT		double

#endif
